/*
 * A quick and dirty program to allow setting of the realtime resource
 * limits under kernel 2.6.12 or later on systems which don't use PAM.
 *
 * Copyright (C) 2005 Jonathan Woithe
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
 * Public License for more details.
 *       
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 675 Mass Ave, Cambridge, MA 02139, USA.
 *       
 */
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>

/* ======================================================================== */

/* Define extended RT limit values if these are for some reason
 * not visible (from taken 2.6.12).
 */
#ifndef RLIMIT_NICE
#define RLIMIT_NICE            13      /* max nice prio allowed to raise to
                                          0-39 for nice level 19 .. -20 */
#define RLIMIT_RTPRIO          14      /* maximum realtime priority */
#endif

/* ======================================================================== */

/* Globals for configuration purposes */

#define VERSION "1.0.0"

/* Location and name of the configuration file.  Normally this will be 
 * something like /etc/set_rtlimits.conf.
 */
#define CONFIG_FILE "/etc/set_rtlimits.conf"

/* Default nice priority (a priority, not a nice level) */
#define DEFAULT_NICE_PRIORITY  20

/* Default RT priority.  Jackd 0.99.73 is happy with 30, but Muse 0.7.2pre1
 * seems to insist on having at least 80.
 */
#define DEFAULT_RT_PRIORITY 80

/* The error code returned if set_rtlimits exits abnormally */
#define ERROR_RETURN_CODE   255

signed int debug = 0;
signed int verbose = 0;
signed int nice_priority = -1;
signed int rt_priority = -1;

/* ======================================================================== */

void show_usage() {

  fprintf(stderr,
    "set_rtlimits - allow non-privileged access to realtime scheduling\n"
    "(c) 2005 Jonathan Woithe.  Released under the terms of the GNU General Public\n"
    "License - no warranty is given or implied.  See source for full details.\n"
    "Usage: set_rtlimits [options] <program> [<program options>]\n"
    "  <program> is the full path to a program to execute after granting realtime\n"
    "  scheduling permission.  <program options> are any options to pass to\n"
    "  <program>.\n"
    "Available options:\n"
    " -d               Turn on additional debugging output\n"
    " -h               Display this help message\n"
    " -n[=<priority>]  Set nice priority limit to <priority> (default %d), where\n"
    "                  0 to 39 equate to nice levels 19 to -20 respectively\n"
    " -r[=<priority>]  Set realtime priority limit to <priority> (default %d)\n"
    " -v               Print version information\n"
    "\n"
    "The maximum realtime priority available is currently 100.\n"
    "Exit code is %d if set_rtlimits encountered an error, or the exit code of\n"
    "the executed program.\n"
    ,DEFAULT_NICE_PRIORITY,DEFAULT_RT_PRIORITY,ERROR_RETURN_CODE); 
}
/* ======================================================================== */

signed int parse_priority(signed int arg_index, int argc, char *argv[], signed int max_priority) {
/* 
 * Parses a priority number from the argument list.  The argument at
 * argv[arg_index] is assumed to be the option which triggers the search
 * for a priority.  Return value is the priority level specified, -1 to
 * indicate that a default value should be used, or -2 to indicate an
 * error.  max_priority allows this function to impose an upper limit on
 * the priority specified (the lower bound is always 0).
 *
 * A priority is specified after an = sign following the option letter.
 */
signed int res;
  /* No priority given, so use default value */
  if (argv[arg_index][2] == 0)
    return -1;

  /* Malformed option: flag an error */
  if (argv[arg_index][2] != '=')
    return -2;

  /* Scan the argument provided and sanity check */
  if (sscanf(argv[arg_index]+3,"%d",&res) != 1) 
    return -2;
  if (res<0 || res>max_priority) {
    fprintf(stderr,"set_rtlimits: priority in `%s' option out of range (min = 0, max = %d)\n",
      argv[arg_index], max_priority);
    return -2;
  }
  return res;
}
/* ======================================================================== */

signed int parse_args(int argc, char *argv[]) {
/*
 * Simplistically parses the command line arguments given.  Returns -1 on
 * error requiring usage information be displayed, -2 for an exit without
 * usage information or the argv element at which the options end (and the
 * program to execvp() starts).
 */
signed int i = 1;
signed int err = 0;

  while (!err && i<argc && argv[i][0]=='-') {
    /* Ensure boolean options are not misformed.  Handlers for options with
     * arguments (optional or otherwise) need to check for correct form
     * themselves.
     */
    if (argv[i][2]!=0 && !strchr("nr",argv[i][1])) {
      err = 1;
    } else {
      switch (argv[i][1]) {
        case 'd': debug = 1; break;
        case 'h': err = -1; break;
        case 'n': 
          nice_priority = parse_priority(i,argc,argv,39);
          if (nice_priority == -1)
            nice_priority = DEFAULT_NICE_PRIORITY;
          else
          if (nice_priority < 0)
            err = -2;
          break;
        case 'r':
          rt_priority = parse_priority(i,argc,argv,100);
          if (rt_priority == -1)
            rt_priority = DEFAULT_RT_PRIORITY;
          else
          if (rt_priority < 0)
            err = -2;
          break;
        case 'v':
          fprintf(stderr,
            "set_rtlimits " VERSION "\n"
          );
          err = -2;
          break;
        default:
          err = -1;
      }
    }
    if (!err) 
      i++;
  }
  if (err && err!=-2 && argv[i][1]!='h') {
    fprintf(stderr,"set_rtlimit: unknown/erroneous option `%s'\n",argv[i]);
  }
  return err?err:i;
}
/* ======================================================================== */

char *skip_whitespace(char *c) {
/*
 * Skips leading whitespace in c and returns a pointer to the first character
 * after the whitespace.  If there are no non-whitespace characters, NULL
 * is returned.
 */
char *p = c;
  if (c==NULL)
    return NULL;
  while (*p!=0 && (*p==' ' || *p=='\t'))
    p++;
  if (*p==0)
    return NULL;
  return p;
}
/* ======================================================================== */

char *skip_to_whitespace(char *c) {
/*
 * Skips to the end of a non-whitespace sequence and returns a pointer to
 * the first whitespace character.  If there is no whitespace to be found,
 * NULL is returned.
 */
char *p = c;
  if (c==NULL)
    return NULL;
  while (*p!=0 && *p!=' ' && *p!='\t')
    p++;
  if (*p==0)
    return NULL;
  return p;
}
/* ======================================================================== */

signed int check_config_file(uid_t uid, gid_t gid, char *exec_path) {
/*
 * Checks to see if the current user has permission to execute the requested
 * program with the specified resource limits.  This information is
 * obtained from a fixed configuration file CONFIG_FILE.
 *
 * The configuration file is a simple textfile with lines of the form
 *   user/group  program  max_nice_priority  max_rt_priority
 * user/group can be ALL which matches all users.  If a user and group has
 * the same name, the match will be against the user.  user/group is
 * expected to be a user/group name.  Where a given user could match multiple
 * entries, the first matching entry encountered is acted upon.  If either
 * priority is <0 it effectively disallows any setting of that resource
 * limit when the entry is matched.
 *
 * Return value is 1 if everything is OK, 0 if the configuration file denies
 * access (by virtue of there being no matching entires or requested limits
 * being out of bounds), or -1 in case of error.  Error messages are generated
 * within this function, so there's no real need for the caller to know
 * the true cause of the error.
 */

FILE *cf_file;
char buf[1024], *c;
char *user, *prog, *max_nice, *max_rt;
uid_t uid_from_file = -1;
gid_t gid_from_file = -1;
signed int match = 0;
signed int err = 0;
signed int line = 0;
signed int max_nice_priority, max_rt_priority;

struct passwd *pwbuf;
struct group *gbuf;

  /* Open configuration file and see if there's a matching entry */
  cf_file = fopen(CONFIG_FILE,"r");
  if (!cf_file) {
    fprintf(stderr,"set_rtlimits: could not access configuration file %s to verify permissions\n",
      CONFIG_FILE);
    return -1;
  }
  do {
    line++;
    buf[0] = 0;
    fgets(buf,1024,cf_file);
    if (!strchr(buf,'\n') && !feof(cf_file)) {
      fprintf(stderr,"set_rtlimits: line %d in %s is too long (maximum length is 1024 characters)\n",
        line, CONFIG_FILE);
      err = 1;
      continue;
    }
    /* Deal with newlines, comment characters, and then split the line into
     * its components
     */
    if ((c=strchr(buf,'\n')) != NULL)
      *c = 0;
    if ((c=strchr(buf,'#')) != NULL)
      *c = 0;
    if (buf[0] == 0)
      continue;
    user = skip_whitespace(buf);
    if ((c=skip_to_whitespace(user))!=NULL)
      *(c++) = 0;
    prog = skip_whitespace(c);
    if ((c=skip_to_whitespace(prog))!=NULL)
      *(c++) = 0;
    max_nice = skip_whitespace(c);
    if ((c=skip_to_whitespace(max_nice))!=NULL)
      *(c++) = 0;
    max_rt = skip_whitespace(c);
    if ((c=skip_to_whitespace(max_rt))!=NULL)
      *c = 0;

    /* If the line is in error, flag the error and abort */
    if (!user || !prog || !max_nice || !max_rt) {
      fprintf(stderr,"set_rtlimits: line %d in %s is not complete\n",
        line, CONFIG_FILE);
      err = 1;
      continue;
    }
    if (sscanf(max_nice,"%d",&max_nice_priority)!=1 ||
        sscanf(max_rt,"%d",&max_rt_priority)!=1) {
      fprintf(stderr,"set_rtlimits: priorities on line %d in %s corrupt\n",
        line, CONFIG_FILE);
      err = 1;
      continue;
    }
    if (prog[0] != '/') {
      fprintf(stderr,"set_rtlimits: warning: program on line %d in %s does not include an absolute path\n",
        line, CONFIG_FILE);
    }

    /* See if user/group in line matches current user.  Do this before the
     * trivial test for a program name match so the accompanying warning is
     * seen even if the line doesn't apply in this case.
     */
    if (!strcmp(user,"ALL")) {
      if (debug) {
        fprintf(stderr,"ALL user/group match detected\n");
      }
      uid_from_file = uid;
    } else
    if ((pwbuf=getpwnam(user)) == NULL) {
      gbuf = getgrnam(user);
      if (gbuf != NULL)
        gid_from_file = gbuf->gr_gid;
    } else
      uid_from_file = pwbuf->pw_uid;
    if (uid_from_file==-1 && gid_from_file==-1) {
      fprintf(stderr,"set_rtlimits: warning: user/group %s on line %d of %s is not a user or group\n",
        user, line, CONFIG_FILE);
      continue;
    }

    /* Skip line if the program name doesn't match */
    if (strcmp(exec_path,prog))
      continue;

    match = (uid_from_file!=-1 && uid_from_file==uid) || (gid_from_file!=-1 && gid_from_file==gid);
  } while (!err && !match && !feof(cf_file));
  fclose(cf_file);
  /* Flag error in configuration file */
  if (err) {
    fprintf(stderr,"set_rtlimits: error in configuration file %s at line %d\n",
      CONFIG_FILE,line);
    return -1;
  }
  /* No matching entry - permission denied */
  if (!match) {
    fprintf(stderr,"set_rtlimits: no permission to run %s with realtime priorities\n",
      exec_path);
    return 0;
  }

  if (debug) {
    fprintf(stderr,"Matching config file entry: uid=%d, gid=%d, prog=%s, max_nice=%d, max_rt=%d\n",
      uid_from_file,gid_from_file,exec_path,max_nice_priority,max_rt_priority);
  }

  /* Check maximum allowed priorities given in configuration file entry */
  if (nice_priority>=0 && nice_priority>max_nice_priority) {
    fprintf(stderr,"set_rtlimits: Requested nice priority limit (%d) not permitted for %s (maximum allowed is %d)\n",
      nice_priority, exec_path, max_nice_priority);
    return 0;
  }
  if (rt_priority>=0 && rt_priority>max_rt_priority) {
    fprintf(stderr,"set_rtlimits: Requested realtime priority limit (%d) not permitted for %s (maximum allowed is %d)\n",
      rt_priority, exec_path, max_rt_priority);
    return 0;
  }

  /* Everything appears in order, so give the OK */
  return 1;
}
/* ======================================================================== */

int main(int argc, char *argv[]) {

struct rlimit rlim;
signed int result, prog_argv_index;
uid_t uid, euid;
gid_t gid, egid;

  uid = getuid();
  gid = getgid();
  euid = geteuid();
  egid = getegid();

  if (uid!=0 && euid!=0) {
    fprintf(stderr,"set_rtlimits: warning: not running setuid root, cannot set RT priority limits\n"
      "  Currently running with uid=%d, euid=%d\n",uid,euid);
  }

  /* Drop root privileges for the moment */
  setregid(egid, gid);
  setreuid(euid, uid);

  /* Work out what to do and handle usage errors */
  prog_argv_index = parse_args(argc, argv);
  if (prog_argv_index < 0) {
    if (prog_argv_index == -1) 
      show_usage();
    return ERROR_RETURN_CODE;
  }
  if (prog_argv_index == argc) {
    fprintf(stderr,"set_rtlimits: no specified program to execute - nothing to do\n");
    return ERROR_RETURN_CODE;
  }
  if (argv[prog_argv_index][0] != '/') {
    fprintf(stderr,"set_rtlimits: program to execute does not include absolute path\n");
    return ERROR_RETURN_CODE;
  }
  if (nice_priority<0 && rt_priority<0) {
    fprintf(stderr,"set_rtlimits: warning: no nice or realtime priority limit changes requested;\n"
                   "  running set_rtlimits will have no effect.\n");
  }

  result = check_config_file(uid,gid,argv[prog_argv_index]);
  if (result != 1) {
    return ERROR_RETURN_CODE;
  }

  /* Debug: explore the current RT resource limit settings */
  if (debug) {
    result = getrlimit(RLIMIT_NICE,&rlim);
    fprintf(stderr,"NICE getrlimit returned %d;   ",result);
    fprintf(stderr,"current NICE rlimits: soft=%u, hard=%u\n",
      (unsigned int)rlim.rlim_cur,(unsigned int)rlim.rlim_max);
    result = getrlimit(RLIMIT_RTPRIO,&rlim);
    fprintf(stderr,"RTPRIO getrlimit returned %d; ",result);
    fprintf(stderr,"current RTPRIO rlimits: soft=%u, hard=%u\n",
      (unsigned int)rlim.rlim_cur,(unsigned int)rlim.rlim_max);
  }

  /* Regain root privileges to set the resource limits */
  setregid(gid,egid);
  setreuid(uid,euid);

  /* Now try to set the NICE resource limit if requested */
  if (nice_priority >= 0) {
    rlim.rlim_max = rlim.rlim_cur = nice_priority;
    result = setrlimit(RLIMIT_NICE,&rlim);
    if (result<0) {
      fprintf(stderr,"set_rtlimits: error setting nice priority resource limit: %s\n"
                     "  Your kernel probably doesn't have RLIMIT_NICE support; try 2.6.12 or later\n",
        strerror(errno));
    }
    if (debug) {
      fprintf(stderr,"NICE setrlimit returned %d",result);
      if (result<0) {
        fprintf(stderr,": errno = %d\n",errno);
      }
      fprintf(stderr,"\n");
    }
  }

  /* Now try to set the RTPRIO resource limit if requested */
  if (rt_priority >= 0) {
    rlim.rlim_max = rlim.rlim_cur = rt_priority;
    result = setrlimit(RLIMIT_RTPRIO,&rlim);
    if (result<0) {
      fprintf(stderr,"set_rtlimits: error setting realtime priority resource limit: %s\n"
                     "  Your kernel probably doesn't have RLIMIT_RTPRIO support; try 2.6.12 or later\n",
        strerror(errno));
    }
    if (debug) {
      fprintf(stderr,"RTPRIO setrlimit returned %d",result);
      if (result<0) {
        fprintf(stderr,": errno = %d\n",errno);
      }
      fprintf(stderr,"\n");
    }
  }

  /* Drop root privileges for good */
  setregid(gid, gid);
  setreuid(uid, uid);

  /* Finally, report final RT resource limits */
  if (debug) {
    result = getrlimit(RLIMIT_NICE,&rlim);
    fprintf(stderr,"NICE getrlimit returned %d; ",result);
    fprintf(stderr,"new NICE rlimits: soft=%u, hard=%u\n",
      (unsigned int)rlim.rlim_cur,(unsigned int)rlim.rlim_max);
    result = getrlimit(RLIMIT_RTPRIO,&rlim);
    fprintf(stderr,"RTPRIO getrlimit returned %d; ",result);
    fprintf(stderr,"new RTPRIO rlimits: soft=%u, hard=%u\n",
      (unsigned int)rlim.rlim_cur,(unsigned int)rlim.rlim_max);
  }

  /* Now execute the main target application with its parameters */
  execvp(argv[prog_argv_index], argv+prog_argv_index);

  fprintf(stderr,"set_rtlimits: could not exec %s: %s\n",
    argv[prog_argv_index],strerror(errno));

  return ERROR_RETURN_CODE;
}
